"""
Quickbooks Online Tax Codes
"""
from .api_base import ApiBase


class TaxCodes(ApiBase):
    """Class for TaxCode APIs."""

    GET_TAX_CODES = '/query?query=select * from TaxCode STARTPOSITION {0} MAXRESULTS 1000'
    COUNT_TAX_CODES = '/query?query=select count(*) from TaxCode where Active = True'

    def get(self):
        """Get a list of the existing Tax Code in the Organization.

        Returns:
            Dict in TaxCode schema.
        """
        return self._query_get_all('TaxCode', TaxCodes.GET_TAX_CODES)

    def get_all_generator(self):
        """Get a list of the existing Tax Code in the Organization.

        Returns:
            Generator with dicts in TaxCode schema.
        """
        return self._query_get_all_generator('TaxCode', TaxCodes.GET_TAX_CODES)

    def count(self):
        """Get count of Tax codes in the Organization.

        Returns:
            Count in Int.
        """
        return self._query(TaxCodes.COUNT_TAX_CODES)['totalCount']



import logging
from test.common.utilities import dict_compare_keys

logger = logging.getLogger(__name__)


def test_accounts(qbo, mock_qbo):
    """
    Test Quickbooks accounts Object
    :param qbo: qbo sdk instance
    :param mock_qbo: mock instance
    :return: None
    """
    accounts = qbo.accounts.get()
    mock_accounts = mock_qbo.accounts.get()

    assert dict_compare_keys(accounts[0], mock_accounts[0]) == [], 'qbo.accounts.get() has stuff that mock_qbo doesnt'
    assert dict_compare_keys(mock_accounts[0], accounts[0]) == [], 'mock_qbo.accounts.get() has stuff that qbo doesnt'


def test_departments(qbo, mock_qbo):
    """
    Test Quickbooks departments Object
    :param qbo: qbo sdk instance
    :param mock_qbo: mock instance
    :return: None
    """
    departments = qbo.departments.get()
    mock_departments = mock_qbo.departments.get()

    assert dict_compare_keys(departments[0], mock_departments[0]) == [], \
        'qbo.departments.get() has stuff that mock_qbo doesnt'
    assert dict_compare_keys(mock_departments[0], departments[0]) == [], \
        'mock_qbo.departments.get() has stuff that qbo doesnt'
    

def test_employees(qbo, mock_qbo):
    """
    Test Quickbooks employees Object
    :param qbo: qbo sdk instance
    :param mock_qbo: mock instance
    :return: None
    """
    employees = qbo.employees.get()
    mock_employees = mock_qbo.employees.get()

    assert dict_compare_keys(employees[0], mock_employees[0]) == [], \
        'qbo.employees.get() has stuff that mock_qbo doesnt'
    assert dict_compare_keys(mock_employees[0], employees[0]) == [], \
        'mock_qbo.employees.get() has stuff that qbo doesnt'


def test_classes(qbo, mock_qbo):
    """
    Test Quickbooks classes Object
    :param qbo: qbo sdk instance
    :param mock_qbo: mock instance
    :return: None
    """
    classes = qbo.classes.get()
    mock_classes = mock_qbo.classes.get()

    assert dict_compare_keys(classes[0], mock_classes[0]) == [], \
        'qbo.classes.get() has stuff that mock_qbo doesnt'
    assert dict_compare_keys(mock_classes[0], classes[0]) == [], \
        'mock_qbo.classes.get() has stuff that qbo doesnt'


def test_exchange_rates(qbo, mock_qbo):
    """
    Test Quickbooks exchange_rates Object
    :param qbo: qbo sdk instance
    :param mock_qbo: mock instance
    :return: None
    """
    exchange_rates = qbo.exchange_rates.get()
    mock_exchange_rates = mock_qbo.exchange_rates.get()

    assert dict_compare_keys(exchange_rates[0], mock_exchange_rates[0]) == [], \
        'qbo.exchange_rates.get() has stuff that mock_qbo doesnt'
    assert dict_compare_keys(mock_exchange_rates[0], exchange_rates[0]) == [], \
        'mock_qbo.exchange_rates.get() has stuff that qbo doesnt'


def test_preferences(qbo, mock_qbo):
    """
    Test Quickbooks preferences Object
    :param qbo: qbo sdk instance
    :param mock_qbo: mock instance
    :return: None
    """
    preferences = qbo.preferences.get()
    mock_preferences = mock_qbo.preferences.get()

    assert dict_compare_keys(preferences, mock_preferences) == [], \
        'qbo.preferences.get() has stuff that mock_qbo doesnt'
    assert dict_compare_keys(mock_preferences, preferences) == [], \
        'mock_qbo.preferences.get() has stuff that qbo doesnt'


def test_journal_entries(qbo, mock_qbo):
    """
    Test Quickbooks journal_entries Object
    :param qbo: qbo sdk instance
    :param mock_qbo: mock instance
    :return: None
    """
    journal_entries = qbo.journal_entries.get()
    mock_journal_entries = mock_qbo.journal_entries.get()

    assert dict_compare_keys(journal_entries[0], mock_journal_entries[0]) == [], \
        'qbo.journal_entries.get() has stuff that mock_qbo doesnt'
    assert dict_compare_keys(mock_journal_entries[0], journal_entries[0]) == [], \
        'mock_qbo.journal_entries.get() has stuff that qbo doesnt'


def test_purchases(qbo, mock_qbo):
    """
    Test Quickbooks purchases Object
    :param qbo: qbo sdk instance
    :param mock_qbo: mock instance
    :return: None
    """
    purchases = qbo.purchases.get()
    mock_purchases = mock_qbo.purchases.get()

    assert dict_compare_keys(purchases[0], mock_purchases[0]) == [], \
        'qbo.purchases.get() has stuff that mock_qbo doesnt'
    assert dict_compare_keys(mock_purchases[0], purchases[0]) == [], \
        'mock_qbo.purchases.get() has stuff that qbo doesnt'


def test_attachments(qbo, mock_qbo):
    """
    Test Quickbooks attachments Object
    :param qbo: qbo sdk instance
    :param mock_qbo: mock instance
    :return: None
    """
    attachments = qbo.attachments.get()
    mock_attachments = mock_qbo.attachments.get()

    assert dict_compare_keys(attachments[0], mock_attachments[0]) == [], \
        'qbo.attachments.get() has stuff that mock_qbo doesnt'
    assert dict_compare_keys(mock_attachments[0], attachments[0]) == [], \
        'mock_qbo.attachments.get() has stuff that qbo doesnt'



import json
import logging
from os import path
from unittest.mock import Mock

from qbosdk import QuickbooksOnlineSDK
logger = logging.getLogger(__name__)


def get_mock_qbo_dict(filename):
    basepath = path.dirname(__file__)
    filepath = path.join(basepath, filename)
    mock_qbo_json = open(filepath, 'r').read()
    mock_qbo_dict = json.loads(mock_qbo_json)
    return mock_qbo_dict


def get_mock_qbo_from_file(filename):
    mock_qbo_dict = get_mock_qbo_dict(filename)
    mock_qbo = Mock()
    mock_qbo.accounts.get.return_value = mock_qbo_dict['accounts']
    mock_qbo.employees.get.return_value = mock_qbo_dict['employees']
    mock_qbo.departments.get.return_value = mock_qbo_dict['departments']
    mock_qbo.classes.get.return_value = mock_qbo_dict['classes']
    mock_qbo.exchange_rates.get.return_value = mock_qbo_dict['exchange_rates']
    mock_qbo.preferences.get.return_value = mock_qbo_dict['preferences']
    mock_qbo.journal_entries.get.return_value = mock_qbo_dict['journal_entries']
    mock_qbo.purchases.get.return_value = mock_qbo_dict['purchases']
    mock_qbo.attachments.get.return_value = mock_qbo_dict['attachments']
    return mock_qbo


def get_mock_qbo():
    return get_mock_qbo_from_file('mock_qbo.json')


def dict_compare_keys(d1, d2, key_path=''):
    """
    Compare two dicts recursively and see if dict1 has any keys that dict2 does not
    Returns: list of key paths
    """
    res = []
    if not d1:
        return res
    if not isinstance(d1, dict):
        return res
    for k in d1:
        if k not in d2:
            missing_key_path = f'{key_path}->{k}'
            res.append(missing_key_path)
        else:
            if isinstance(d1[k], dict):
                key_path1 = f'{key_path}->{k}'
                res1 = dict_compare_keys(d1[k], d2[k], key_path1)
                res = res + res1
            elif isinstance(d1[k], list):
                key_path1 = f'{key_path}->{k}[0]'
                dv1 = d1[k][0] if len(d1[k]) > 0 else None
                dv2 = d2[k][0] if len(d2[k]) > 0 else None
                res1 = dict_compare_keys(dv1, dv2, key_path1)
                res = res + res1
    return res


def qbo_connect():
    """
    QBO Connector connection
    :param: db connection
    """
    file = open('test_credentials.json', 'r')
    quickbooks_config = json.load(file)

    connection = QuickbooksOnlineSDK(
        client_id=quickbooks_config['client_id'],
        client_secret=quickbooks_config['client_secret'],
        refresh_token=quickbooks_config['refresh_token'],
        realm_id=quickbooks_config['realm_id'],
        environment=quickbooks_config['environment']
    )

    quickbooks_config['refresh_token'] = connection.refresh_token

    with open('test_credentials.json', 'w') as fp:
        json.dump(quickbooks_config, fp)
    return connection



"""
Quickbooks Online journal entries
"""
from typing import Dict
from .api_base import ApiBase


class JournalEntries(ApiBase):
    """Class for Categories APIs."""

    GET_JOURNAL_ENTRIES = '/query?query=select * from JournalEntry STARTPOSITION {0} MAXRESULTS 1000'
    POST_JOURNAL_ENTRY = '/journalentry?minorversion=53'
    DELETE_JOURNAL_ENTRY = '/journalentry?operation=delete'
    GET_JOURNAL_ENTRY_BY_ID = '/journalentry/{0}'

    def get(self):
        """
        Get all JournalEntries
        :return: List of Dicts in JournalEntry Schema
        """
        return self._query_get_all('JournalEntry', JournalEntries.GET_JOURNAL_ENTRIES)

    def get_all_generator(self):
        """
        Get all JournalEntries
        :return: Generator with Dicts in JournalEntry Schema
        """
        return self._query_get_all_generator('JournalEntry', JournalEntries.GET_JOURNAL_ENTRIES)

    def post(self, data: Dict):
        """
        Post JournalEntry to Quickbooks Online
        :param data: Dict in JournalEntry schema
        :return:
        """
        return self._post_request(data, JournalEntries.POST_JOURNAL_ENTRY)

    def get_by_id(self, journal_entry_id):
        """
        Get JournalEntry by Id
        :param journal_entry_id: Journal Entry Id
        :return: Dict in JournalEntry schema
        """
        return self._get_request('JournalEntry', JournalEntries.GET_JOURNAL_ENTRY_BY_ID.format(journal_entry_id))

    def delete(self, journal_entry_id: str):
        """
        Delete JournalEntry from Quickbooks Online
        :param journal_entry_id: Journal Entry Id to remove
        :return: Dict response
        """
        data = {
            'Id': journal_entry_id,
            'SyncToken': '1'
        }
        return self._post_request(data, JournalEntries.DELETE_JOURNAL_ENTRY)



"""
Quickbooks Online TaxRates
"""
from .api_base import ApiBase


class TaxRates(ApiBase):
    """Class for TaxRates APIs."""

    GET_TAX_RATES = '/query?query=select * from TaxRate STARTPOSITION {0} MAXRESULTS 1000'
    GET_TAX_RATES_BY_ID = '/taxrate/{0}'

    def get(self):
        """
        Get all Taxrates
        :return: List of Dicts in Taxrates Schema
        """
        return self._query_get_all('TaxRate', TaxRates.GET_TAX_RATES)

    def get_all_generator(self):
        """
        Get all Taxrates
        :return: Generator of Dicts in Taxrates Schema
        """
        return self._query_get_all_generator('TaxRate', TaxRates.GET_TAX_RATES)

    def get_by_id(self, taxrateId: str):
        """
        Get Taxrates from Quickbooks Online
        :param taxrateId: Taxrate Id
        :return: Dict Response
        """
        return self._get_request('TaxRate', TaxRates.GET_TAX_RATES_BY_ID.format(taxrateId))



import logging
import pytest

from test.common.utilities import get_mock_qbo, qbo_connect

logger = logging.getLogger(__name__)


@pytest.fixture
def mock_qbo():
    return get_mock_qbo()


@pytest.fixture(scope='module')
def qbo():
    return qbo_connect()





"""
Quickbooks Online accounts
"""
from .api_base import ApiBase


class Accounts(ApiBase):
    """Class for Categories APIs."""

    GET_ACCOUNTS = '/query?query=select * from Account STARTPOSITION {0} MAXRESULTS 1000'
    ACCOUNT_COUNT = '/query?query=select count(*) from Account where Active = True'

    def get(self):
        """Get a list of the existing Accounts in the Organization.

        Returns:
            List with dicts in Accounts schema.
        """
        return self._query_get_all('Account', Accounts.GET_ACCOUNTS)

    def get_all_generator(self):
        """Get a generator of all the existing Accounts in the Organization.

        Returns:
            Generator with dicts in Accounts schema.
        """
        return self._query_get_all_generator('Account', Accounts.GET_ACCOUNTS)

    def get_inactive(self, last_updated_time: None):
        """
        Retrieves a list of inactive accounts from the QuickBooks Online API.

        :param last_updated_time: The last updated time to filter the accounts.
        :return: A list of inactive accounts.
        """

        QUERY = "/query?query=select * from Account where Active=false"
        if last_updated_time:
            QUERY += f" and Metadata.LastUpdatedTime >= '{last_updated_time}'"
        QUERY += " STARTPOSITION {0} MAXRESULTS 1000"

        return self._query_get_all_generator('Account', QUERY)

    def count(self):
        """Get count of Accounts in the Organization.

        Returns:
            Count in Int.
        """
        return self._query(Accounts.ACCOUNT_COUNT)['totalCount']



"""
Quickbooks Online vendors
"""
from typing import Dict

from .api_base import ApiBase


class Vendors(ApiBase):
    """Class for Categories APIs."""

    GET_VENDORS = '/query?query=select * from Vendor STARTPOSITION {0} MAXRESULTS 1000'
    POST_VENDOR = '/vendor?minorversion=38'
    SEARCH_VENDOR = "/query?query=select * from Vendor where DisplayName = '{0}'"
    COUNT_VENDORS = '/query?query=select count(*) from Vendor where Active = True'

    def get(self):
        """Get a list of the existing Vendors in the Organization.

        Returns:
            List with dicts in Vendors schema.
        """
        return self._query_get_all('Vendor', Vendors.GET_VENDORS)

    def get_all_generator(self):
        """Get a list of the existing Vendors in the Organization.

        Returns:
            Generator with dicts in Vendors schema.
        """
        return self._query_get_all_generator('Vendor', Vendors.GET_VENDORS)

    def post(self, data: Dict):
        """
        Post Vendor to Quickbooks Online
        :param data: Dict in Vendor schema
        :return:
        """
        return self._post_request(data, Vendors.POST_VENDOR)

    def search_vendor_by_display_name(self, display_name=None):
        """
        Search vendor by display name
        :param display_name: Quickbooks Vendor Display Name
        :return: Vendor
        """
        response = self._get_request('QueryResponse', Vendors.SEARCH_VENDOR.format(display_name))

        return response['Vendor'][0] if 'Vendor' in response else None

    def get_inactive(self, last_updated_time: None):
        """
        Retrieves a list of inactive vendors from the QuickBooks Online API.

        :param last_updated_time: The last updated time to filter the vendors.
        :return: A list of inactive vendors.
        """

        QUERY = "/query?query=select * from Vendor where Active=false"
        if last_updated_time:
            QUERY += f" and Metadata.LastUpdatedTime >= '{last_updated_time}'"
        QUERY += " STARTPOSITION {0} MAXRESULTS 1000"

        return self._query_get_all_generator('Vendor', QUERY)

    def count(self):
        """Get count of Vendors in the Organization.

        Returns:
            Count in Int.
        """
        return self._query(Vendors.COUNT_VENDORS)['totalCount']



"""
Quickbooks Online preferences
"""
from .api_base import ApiBase


class CompanyInfo(ApiBase):
    """Class for Company Info APIs."""

    GET_COMPANY_INFO = '/companyinfo/{0}'

    def __init__(self, realm_id):
        super().__init__()
        self.__realm_id = realm_id

    def get(self):
        """Get a list of the existing Preferences in the Organization.

        Returns:
            Dict in Preferences schema.
        """
        return self._get_request('CompanyInfo', CompanyInfo.GET_COMPANY_INFO.format(self.__realm_id))



"""
Quickbooks Online init
"""
from .qbosdk import QuickbooksOnlineSDK, revoke_refresh_token
from.exceptions import *

__all__ = [
    'QuickbooksOnlineSDK',
    'QuickbooksOnlineSDKError',
    'NotFoundClientError',
    'UnauthorizedClientError',
    'ExpiredTokenError',
    'InvalidTokenError',
    'NoPrivilegeError',
    'WrongParamsError',
    'NotFoundItemError',
    'InternalServerError',
    'revoke_refresh_token'
]

name = "qbosdk"



"""
Quickbooks Online purchases
"""
from typing import Dict
from .api_base import ApiBase


class Purchases(ApiBase):
    """Class for Categories APIs."""

    GET_PURCHASES = '/query?query=select * from Purchase STARTPOSITION {0} MAXRESULTS 1000'
    POST_PURCHASE = '/purchase?minorversion=38'
    DELETE_PURCHASE = '/purchase?operation=delete'
    GET_PURCHASE_BY_ID = "/query?query=select * from Purchase where Id = \'{0}\'"

    def get(self):
        """
        Get all Purchases
        :return: List of Dicts in Purchase Schema
        """
        return self._query_get_all('Purchase', Purchases.GET_PURCHASES)

    def get_all_generator(self):
        """
        Get all Purchases
        :return: Generator with Dicts in Purchase Schema
        """
        return self._query_get_all_generator('Purchase', Purchases.GET_PURCHASES)

    def post(self, data: Dict):
        """
        Post Purchase (check, etc) to Quickbooks Online
        :param data: Dict in Purchase schema
        :return:
        """
        return self._post_request(data, Purchases.POST_PURCHASE)

    def delete(self, purchase_id: str):
        """
        Delete Purchase (check, etc) from Quickbooks Online
        :param purchase_id: Journal Entry Id to remove
        :return: Dict response
        """
        data = {
            'Id': purchase_id,
            'SyncToken': '1'
        }
        return self._post_request(data, Purchases.DELETE_PURCHASE)

    def get_by_id(self, purchase_id: str):
        """
        Get Purchase by Id
        :param purchase_id: Purchase Id
        :return: Dict in Purchase schema
        """
        return self._query(self.GET_PURCHASE_BY_ID.format(purchase_id))



"""
Quickbooks Online classes
"""
from .api_base import ApiBase


class Classes(ApiBase):
    """Class for Categories APIs."""

    GET_CLASSES = '/query?query=select * from Class STARTPOSITION {0} MAXRESULTS 1000'
    COUNT_CLASSES = '/query?query=select count(*) from Class where Active = True'

    def get(self):
        """Get a list of the existing Classes in the Organization.

        Returns:
            List with dicts in Classes schema.
        """
        return self._query_get_all('Class', Classes.GET_CLASSES)

    def get_all_generator(self):
        """Get a list of the existing Classes in the Organization.

        Returns:
            Generator with dicts in Classes schema.
        """
        return self._query_get_all_generator('Class', Classes.GET_CLASSES)


    def count(self):
        """Get count of Classes in the Organization.

        Returns:
            Count in Int.
        """
        return self._query(Classes.COUNT_CLASSES)['totalCount']



"""
Quickbooks Online vendors
"""
from .api_base import ApiBase


class Customers(ApiBase):
    """Class for Customer APIs."""

    GET_CUSTOMERS = '/query?query=select * from Customer STARTPOSITION {0} MAXRESULTS 1000'
    COUNT_CUSTOMERS = '/query?query=select count(*) from Customer where Active = True'

    def get(self):
        """Get a list of the existing Customers in the Organization.

        Returns:
            List with dicts in Customers schema.
        """
        return self._query_get_all('Customer', Customers.GET_CUSTOMERS)

    def get_all_generator(self):
        """Get a list of the existing Customers in the Organization.

        Returns:
            Generator with dicts in Customers schema.
        """
        return self._query_get_all_generator('Customer', Customers.GET_CUSTOMERS)

    def count(self):
        """Get count of Customers in the Organization.

        Returns:
            Count in Int.
        """
        return self._query(Customers.COUNT_CUSTOMERS)['totalCount']

    def get_inactive(self, last_updated_time: None):
        """
        Retrieves a list of inactive customers from the QuickBooks Online API.

        :param last_updated_time: The last updated time to filter the customers.
        :return: A list of inactive customers.
        """

        QUERY = "/query?query=select * from Customer where Active=false"
        if last_updated_time:
            QUERY += f" and Metadata.LastUpdatedTime >= '{last_updated_time}'"
        QUERY += " STARTPOSITION {0} MAXRESULTS 1000"

        return self._query_get_all_generator('Customer', QUERY)



"""
API Base class with util functions
"""
import logging
import json
from typing import List, Dict, Generator

import requests

from ..exceptions import WrongParamsError, InvalidTokenError, QuickbooksOnlineSDKError, \
    NoPrivilegeError, NotFoundItemError, ExpiredTokenError, InternalServerError

logger = logging.getLogger(__name__)

class ApiBase:
    """The base class for all API classes."""

    def __init__(self):
        self.__access_token = None
        self.__server_url = None

    def change_access_token(self, access_token):
        """
        Sets the access token for APIs
        :param access_token: acceess token (JWT)
        :return: None
        """
        self.__access_token = access_token

    def set_server_url(self, server_url: str):
        """
        Set the server url for APIs
        :param server_url: Url with realm id
        :return: None
        """
        self.__server_url = server_url

    def _query_get_all(self, object_type: str, url: str) -> List[Dict]:
        """
        Gets all the objects of a particular type for query type GET calls
        :param url: GET URL of object
        :param object_type: type of object
        :return: list of objects
        """
        start_position = 1

        request_url = '{0}{1}'.format(self.__server_url, url)

        objects = []

        api_headers = {
            'Authorization': 'Bearer {0}'.format(self.__access_token),
            'Accept': 'application/json'
        }

        response = requests.get(url=request_url.format(start_position), headers=api_headers)

        if response.status_code == 200:
            logger.debug('Response for get request for url: %s, %s', url, response.text)
            data = json.loads(response.text)
            query_response = data['QueryResponse']

            while query_response:
                objects.extend(query_response[object_type])
                start_position = start_position + 1000

                data = json.loads(requests.get(url=request_url.format(start_position), headers=api_headers).text)

                query_response = data['QueryResponse']
            return objects

        logger.info('Response for get request for url: %s, %s', url, response.text)
        if response.status_code == 400:
            raise WrongParamsError('Some of the parameters are wrong', response.text)

        if response.status_code == 401:
            raise InvalidTokenError('Invalid token, try to refresh it', response.text)

        if response.status_code == 403:
            raise NoPrivilegeError('Forbidden, the user has insufficient privilege', response.text)

        if response.status_code == 404:
            raise NotFoundItemError('Not found item with ID', response.text)

        if response.status_code == 498:
            raise ExpiredTokenError('Expired token, try to refresh it', response.text)

        if response.status_code == 500:
            raise InternalServerError('Internal server error', response.text)

        raise QuickbooksOnlineSDKError('Error: {0}'.format(response.status_code), response.text)

    def _query_get_all_generator(self, object_type: str, url: str) -> Generator[Dict, None, None]:
        """
        Gets all the objects of a particular type for query type GET calls
        :param url: GET URL of object
        :param object_type: type of object
        :return: list of objects
        """
        start_position = 1

        request_url = '{0}{1}'.format(self.__server_url, url)

        api_headers = {
            'Authorization': 'Bearer {0}'.format(self.__access_token),
            'Accept': 'application/json'
        }

        while True:
            try:
                response = requests.get(url=request_url.format(start_position), headers=api_headers)
                response.raise_for_status()

                data = json.loads(response.text)
                query_response = data['QueryResponse']

                if not query_response or object_type not in query_response:
                    break

                logger.debug('Response for get request for url: %s, %s', url, response.text)
                
                yield query_response[object_type]

                start_position += 1000
            
            except requests.exceptions.HTTPError as err:
                logger.info('Response for get request for url: %s, %s', url, err.response.text)
                if err.response.status_code == 400:
                    raise WrongParamsError('Some of the parameters are wrong', err.response.text)

                if err.response.status_code == 401:
                    raise InvalidTokenError('Invalid token, try to refresh it', err.response.text)

                if err.response.status_code == 403:
                    raise NoPrivilegeError('Forbidden, the user has insufficient privilege', err.response.text)

                if err.response.status_code == 404:
                    raise NotFoundItemError('Not found item with ID', err.response.text)

                if err.response.status_code == 498:
                    raise ExpiredTokenError('Expired token, try to refresh it', err.responseesponse.text)

                if err.response.status_code == 500:
                    raise InternalServerError('Internal server error', err.response.text)

                raise QuickbooksOnlineSDKError('Error: {0}'.format(err.response.status_code), err.response.text)

    def _query(self, url: str) -> List[Dict]:
        """
        Returns results for query type GET calls
        :param object_type: type of object
        :return: dict of the response
        """
        request_url = '{0}{1}'.format(self.__server_url, url)
        api_headers = {
            'Authorization': 'Bearer {0}'.format(self.__access_token),
            'Accept': 'application/json'
        }

        response = requests.get(url=request_url, headers=api_headers)

        if response.status_code == 200:
            logger.debug('Response for get request for url: %s, %s', url, response.text)
            data = json.loads(response.text)
            return data['QueryResponse'] if 'QueryResponse' in data else data

        logger.info('Response for get request for url: %s, %s', url, response.text)
        if response.status_code == 400:
            raise WrongParamsError('Some of the parameters are wrong', response.text)

        if response.status_code == 401:
            raise InvalidTokenError('Invalid token, try to refresh it', response.text)

        if response.status_code == 403:
            raise NoPrivilegeError('Forbidden, the user has insufficient privilege', response.text)

        if response.status_code == 404:
            raise NotFoundItemError('Not found item with ID', response.text)

        if response.status_code == 498:
            raise ExpiredTokenError('Expired token, try to refresh it', response.text)

        if response.status_code == 500:
            raise InternalServerError('Internal server error', response.text)

        raise QuickbooksOnlineSDKError('Error: {0}'.format(response.status_code), response.text)

    def _get_request(self, object_type: str, api_url: str) -> List[Dict] or Dict:
        """Create a HTTP GET request.

        Parameters:
            api_url (str): Url for the wanted API.

        Returns:
            A response from the request (dict).
        """

        api_headers = {
            'Authorization': 'Bearer {0}'.format(self.__access_token),
            'Accept': 'application/json'
        }

        response = requests.get(
            '{0}{1}'.format(self.__server_url, api_url),
            headers=api_headers
        )

        if response.status_code == 200:
            logger.debug('Response for get request for url: %s, %s', api_url, response.text)
            result = json.loads(response.text)
            return result[object_type]

        logger.info('Response for get request for url: %s, %s', api_url, response.text)
        if response.status_code == 400:
            raise WrongParamsError('Some of the parameters are wrong', response.text)

        if response.status_code == 401:
            raise InvalidTokenError('Invalid token, try to refresh it', response.text)

        if response.status_code == 403:
            raise NoPrivilegeError('Forbidden, the user has insufficient privilege', response.text)

        if response.status_code == 404:
            raise NotFoundItemError('Not found item with ID', response.text)

        if response.status_code == 498:
            raise ExpiredTokenError('Expired token, try to refresh it', response.text)

        if response.status_code == 500:
            raise InternalServerError('Internal server error', response.text)

        raise QuickbooksOnlineSDKError('Error: {0}'.format(response.status_code), response.text)

    def _post_request(self, data, api_url):
        """Create a HTTP post request.

        Parameters:
            data (dict): HTTP POST body data for the wanted API.
            api_url (str): Url for the wanted API.

        Returns:
            A response from the request (dict).
        """

        api_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Authorization': 'Bearer {0}'.format(self.__access_token)
        }

        response = requests.post(
            '{0}{1}'.format(self.__server_url, api_url),
            headers=api_headers,
            json=data
        )

        if response.status_code == 200:
            logger.debug('Response for post request: %s', response.text)
            result = json.loads(response.text)
            return result
    
        logger.debug('Payload for post request: %s', data)
        logger.info('Response for post request: %s', response.text)

        if response.status_code == 400:
            raise WrongParamsError('Some of the parameters are wrong', response.text)

        if response.status_code == 401:
            raise InvalidTokenError('Invalid token, try to refresh it', response.text)

        if response.status_code == 403:
            raise NoPrivilegeError('Forbidden, the user has insufficient privilege', response.text)

        if response.status_code == 404:
            raise NotFoundItemError('Not found item with ID', response.text)

        if response.status_code == 498:
            raise ExpiredTokenError('Expired token, try to refresh it', response.text)

        if response.status_code == 500:
            raise InternalServerError('Internal server error', response.text)

        raise QuickbooksOnlineSDKError('Error: {0}'.format(response.status_code), response.text)

    def _post_file(self, data, api_url):
        """Create a HTTP post request.

        Parameters:
            data (str): HTTP POST body data for the wanted API.
            api_url (str): Url for the wanted API.

        Returns:
            A response from the request (dict).
        """

        api_headers = {
            'Content-Type': 'multipart/form-data; boundary={0}'.format('YOjcLaTlykb6OxfYJx4O07j1MweeMFem'),
            'Accept': 'application/json',
            'Connection': 'close',
            'Authorization': 'Bearer {0}'.format(self.__access_token)
        }

        response = requests.post(
            '{0}{1}'.format(self.__server_url, api_url),
            headers=api_headers,
            data=data
        )

        if response.status_code == 200:
            logger.debug('Response for post request: %s', response.text)
            result = json.loads(response.text)
            return result['AttachableResponse'][0]['Attachable']


        logger.debug('Payload for post request: %s', data)
        logger.info('Response for post request: %s', response.text)

        if response.status_code == 400:
            raise WrongParamsError('Some of the parameters are wrong', response.text)

        if response.status_code == 401:
            raise InvalidTokenError('Invalid token, try to refresh it', response.text)

        if response.status_code == 403:
            raise NoPrivilegeError('Forbidden, the user has insufficient privilege', response.text)

        if response.status_code == 404:
            raise NotFoundItemError('Not found item with ID', response.text)

        if response.status_code == 498:
            raise ExpiredTokenError('Expired token, try to refresh it', response.text)

        if response.status_code == 500:
            raise InternalServerError('Internal server error', response.text)

        raise QuickbooksOnlineSDKError('Error: {0}'.format(response.status_code), response.text)



"""
Quickbooks Online Bill Payments
"""
from typing import Dict
from .api_base import ApiBase


class BillPayments(ApiBase):
    """Class for Bill Payments API."""

    GET_BILLS = '/query?query=select * from billpayment STARTPOSITION {0} MAXRESULTS 1000'
    POST_BILL_PAYMENT = '/billpayment?minorversion=38'
    DELETE_BILL = '/billpayment?operation=delete'

    def get(self):
        """
        Get all Bill Payments
        :return: List of Dicts in Bill Payment Schema
        """
        return self._query_get_all('BillPayment', BillPayments.GET_BILLS)

    def get_all_generator(self):
        """
        Get all Bill Payments
        :return: Generator of Dicts in Bill Payment Schema
        """
        return self._query_get_all_generator('BillPayment', BillPayments.GET_BILLS)

    def post(self, data: Dict):
        """
        Post Bill Payment to Quickbooks Online
        :param data: Dict in Bill Payment schema
        :return:
        """
        return self._post_request(data, BillPayments.POST_BILL_PAYMENT)

    def delete(self, bill_payment_id: str):
        """
        Delete Bill Payment from Quickbooks Online
        :param bill_payment_id: Bill Payment Id to remove
        :return: Dict response
        """
        data = {
            'SyncToken': '1',
            'Id': bill_payment_id
        }
        return self._post_request(data, BillPayments.DELETE_BILL)



"""
Quickbooks Online employees
"""

from .api_base import ApiBase


class Items(ApiBase):
    """Class for Items APIs."""

    GET_ITEMS = '/query?query=select * from Item STARTPOSITION {0} MAXRESULTS 1000'
    COUNT_ITEMS = '/query?query=select count(*) from Item where Active = True'

    def get(self):
        """Get a list of the existing items in the Organization.

        Returns:
            List with dicts in Items schema.
        """
        return self._query_get_all('Item', Items.GET_ITEMS)

    def get_all_generator(self):
        """Get a list of the existing items in the Organization.

        Returns:
            Generator with dicts in Items schema.
        """
        return self._query_get_all_generator('Item', Items.GET_ITEMS)

    def get_inactive(self, last_updated_time: None):
        """
        Retrieves a list of inactive items from the QuickBooks Online API.

        :param last_updated_time: The last updated time to filter the items.
        :return: A list of inactive items.
        """

        QUERY = "/query?query=select * from Item where Active=false"
        if last_updated_time:
            QUERY += f" and Metadata.LastUpdatedTime >= '{last_updated_time}'"
        QUERY += " STARTPOSITION {0} MAXRESULTS 1000"

        return self._query_get_all_generator('Item', QUERY)

    def count(self):
        """Get count of Items in the Organization.

        Returns:
            Count in Int.
        """
        return self._query(Items.COUNT_ITEMS)['totalCount']



"""
Quickbooks Online SDK Exceptions
"""


class QuickbooksOnlineSDKError(Exception):
    """The base exception class for QBOSDK.

    Parameters:
        msg (str): Short description of the error.
        response: Error response from the API call.
    """

    def __init__(self, msg, response=None):
        super(QuickbooksOnlineSDKError, self).__init__(msg)
        self.message = msg
        self.response = response

    def __str__(self):
        return repr(self.message)


class NotFoundClientError(QuickbooksOnlineSDKError):
    """Client not found OAuth2 authorization, 404 error."""


class UnauthorizedClientError(QuickbooksOnlineSDKError):
    """Wrong client secret and/or refresh token, 401 error."""


class ExpiredTokenError(QuickbooksOnlineSDKError):
    """Expired (old) access token, 498 error."""


class InvalidTokenError(QuickbooksOnlineSDKError):
    """Wrong/non-existing access token, 401 error."""


class NoPrivilegeError(QuickbooksOnlineSDKError):
    """The user has insufficient privilege, 403 error."""


class WrongParamsError(QuickbooksOnlineSDKError):
    """Some of the parameters (HTTP params or request body) are wrong, 400 error."""


class NotFoundItemError(QuickbooksOnlineSDKError):
    """Not found the item from URL, 404 error."""


class InternalServerError(QuickbooksOnlineSDKError):
    """The rest QBOSDK errors, 500 error."""



"""
Quickbooks Online Bills
"""
from typing import Dict
from .api_base import ApiBase


class Bills(ApiBase):
    """Class for Bill APIs."""

    GET_BILLS = '/query?query=select * from Bill STARTPOSITION {0} MAXRESULTS 1000'
    POST_BILL = '/bill?minorversion=38'
    DELETE_BILL = '/bill?operation=delete'
    GET_BILL_BY_ID = '/bill/{0}'

    def get(self):
        """
        Get all Bills
        :return: List of Dicts in Bill Schema
        """
        return self._query_get_all('Bill', Bills.GET_BILLS)

    def get_all_generator(self):
        """
        Get all Bills
        :return: Generator with Dicts in Bill Schema
        """
        return self._query_get_all_generator('Bill', Bills.GET_BILLS)

    def post(self, data: Dict):
        """
        Post Bill to Quickbooks Online
        :param data: Dict in Bill schema
        :return:
        """
        return self._post_request(data, Bills.POST_BILL)

    def delete(self, bill_id: str):
        """
        Delete Bill from Quickbooks Online
        :param bill_id: Bill Id to remove
        :return: Dict response
        """
        data = {
            'Id': bill_id,
            'SyncToken': '1'
        }
        return self._post_request(data, Bills.DELETE_BILL)

    def get_by_id(self, bill_id):
        """
        Get Bill from Quickbooks Online
        :param bill_id: Bill Id
        :return: Dict Response
        """
        return self._get_request('Bill', Bills.GET_BILL_BY_ID.format(bill_id))



"""
Quickbooks Online departments
"""
from .api_base import ApiBase


class Departments(ApiBase):
    """Class for Categories APIs."""

    GET_DEPARTMENTS = '/query?query=select * from Department STARTPOSITION {0} MAXRESULTS 1000'
    COUNT_DEPARTMENT = '/query?query=select count(*) from Department where Active = True'

    def get(self):
        """Get a list of the existing Departments in the Organization.

        Returns:
            List with dicts in Departments schema.
        """
        return self._query_get_all('Department', Departments.GET_DEPARTMENTS)

    def get_all_generator(self):
        """Get a list of the existing Departments in the Organization.

        Returns:
            Generator with dicts in Departments schema.
        """
        return self._query_get_all_generator('Department', Departments.GET_DEPARTMENTS)

    def count(self):
        """Get count of Departments in the Organization.

        Returns:
            Count in Int.
        """
        return self._query(Departments.COUNT_DEPARTMENT)['totalCount']



"""
Project setup file
"""
import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='qbosdk',
    version='0.20.0',
    author='Shwetabh Kumar',
    author_email='shwetabh.kumar@fyle.in',
    description='Python SDK for accessing Quickbooks Online APIs',
    license='MIT',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['quickbooks-online', 'quickbooks', 'fyle', 'api', 'python', 'sdk'],
    url='https://github.com/fylein/qbo-sdk-py',
    packages=setuptools.find_packages(),
    install_requires=['requests>=2.25.0', 'future==0.18.2'],
    classifiers=[
        'Topic :: Internet :: WWW/HTTP',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ]
)



"""
Quickbooks Online SDK init
"""
from .accounts import Accounts
from .departments import Departments
from .classes import Classes
from .employees import Employees
from .preferences import Preferences
from .company_info import CompanyInfo
from .exchange_rates import ExchangeRates
from .purchases import Purchases
from .journal_entries import JournalEntries
from .vendors import Vendors
from .bills import Bills
from .attachments import Attachments
from .customers import Customers
from .bill_payments import BillPayments
from .tax_codes import TaxCodes
from .tax_rates import TaxRates
from .items import Items

__all_ = [
    'Accounts',
    'Departments',
    'Classes',
    'CompanyInfo',
    'Employees',
    'Preferences',
    'ExchangeRates',
    'Purchases',
    'Vendors',
    'Bills',
    'JournalEntries',
    'Attachments',
    'Customers',
    'BillPayments',
    'TaxCodes',
    'TaxRates',
    'Items'
]



"""
Quickbooks Online employees
"""
from typing import Dict

from .api_base import ApiBase


class Employees(ApiBase):
    """Class for Categories APIs."""

    GET_EMPLOYEES = '/query?query=select * from Employee STARTPOSITION {0} MAXRESULTS 1000'
    POST_EMPLOYEE = '/employee?minorversion=38'
    COUNT_EMPLOYEES = '/query?query=select count(*) from Employee where Active = True'

    def get(self):
        """Get a list of the existing Employees in the Organization.

        Returns:
            List with dicts in Employees schema.
        """
        return self._query_get_all('Employee', Employees.GET_EMPLOYEES)

    def get_all_generator(self):
        """Get a list of the existing Employees in the Organization.

        Returns:
            Generator with dicts in Employees schema.
        """
        return self._query_get_all_generator('Employee', Employees.GET_EMPLOYEES)

    def post(self, data: Dict):
        """
        Post Employee to Quickbooks Online
        :param data: Dict in Employee schema
        :return:
        """
        return self._post_request(data, Employees.POST_EMPLOYEE)

    def count(self):
        """Get count of Employees in the Organization.

        Returns:
            Count in Int.
        """
        return self._query(Employees.COUNT_EMPLOYEES)['totalCount']
